/* -*- mode: c++; c-basic-offset: 4; -*- */
//
//   ChanSearch performs a recursive binary search through frame files of 
//   a specified type to find the change-points in the channel contents of
//   the frames.
//
//   The command line syntax is as follows
//    ChanSearch --start-gps <start> --end-gps <end> --loc-file <file>
//               --cache-dir <direc>

#include "types.hh"
#include "frame_source.hh"
#include "chan_config.hh"
#include "scandir.hh"
#include "fzstream.hh"
#include <string>
#include <iostream>
#include <vector>
#include <unistd.h>

sends::frame_source sends::Source_List;
sends::chan_list    sends::Channel_List;

using namespace std;
using namespace sends;

//======================================  Channel search main function.
inline void
syntax(void) {
    cerr << "The command line syntax is as follows: " << endl;
    cerr << "ChanSearch [--gzip | -z] --start-gps <start> --end-gps <end> \\"
	 << endl;
    cerr << "           --loc-file <file> --frame-type <type> "
	 << "--cache-dir <direc>" << endl;
}

//======================================  Channel search main function.
int
main(int argc, const char* argv[]) {
    
    //----------------------------------  Get the arguments
    string filematch = "/data/archive";
    string frame_type = "H-R";
    string cacheDir = ".";
    unsigned long debug = 0;
    bool gzip = false;
    gps_type start = 0;
    gps_type stop  = chan_entry::source_entry::inf_gps;
    for (int i=1; i<argc; i++) {
	string argi = argv[i];
	if (argi == "--cache-dir" || argi == "-d") {
	    cacheDir = argv[++i];
	}
        else if (argi == "--debug") {
            debug = strtoul(argv[++i], 0, 0);
        }
	else if (argi == "--end-gps" || argi == "-e") {
	    stop = strtoul(argv[++i], 0, 0);
	}
	else if (argi == "--frame-type" || argi == "-f") {
	    frame_type = argv[++i];
	}
	else if (argi == "--gzip" || argi == "-z") {
	    gzip = true;
	}
	else if (argi == "--loc-file" || argi == "-l") {
	    Source_List.read(argv[++i]);
	}
	else if (argi == "--start-gps" || argi == "-s") {
	    start = strtoul(argv[++i], 0, 0);
	}
	else {
	    cerr << "Unrecognized argument: " << argi << endl;
	    syntax();
	    return 1;
	}
    }

    //----------------------------------  Check for writeable cache directory
    if (access(cacheDir.c_str(), W_OK)) {
	cerr << "ChanSearch: Cache directory: " << cacheDir 
	     << " is not writeable." << endl;
        syntax();
        return 3;
    }

    //----------------------------------  Check for valid start-point
    const loc_list*  ll = Source_List.find_first(frame_type);
    if (!ll) {
	cerr << "No valid frames for type: " << frame_type << endl;
	return 2;
    }
    gps_type first_start = ll->front().start();
    if (start < first_start) start = first_start;

    //----------------------------------  Check for valid stop point
    ll = Source_List.find_last(frame_type);
    if (!ll) {
	cerr << "ChanSearch: No valid frames for type: " << frame_type << endl;
	return 2;
    }
    gps_type last_stop = ll->back().stop();
    if (last_stop < stop) stop = last_stop;
    if (stop <= start) {
	cerr << "ChanSearch: No " << frame_type << " frames in time range" 
	     << endl;
	return 2;
    }

    //---------------------------------- get the end-point lists
    chan_config topconf;
    if (debug) topconf.set_debug(debug);
    std::vector<chan_config> stack(32);
    stack.clear();
    topconf.getChanList(frame_type, start);
    stack.push_back(topconf);

    frame_loc fl;
    if (!Source_List.find(frame_type, stop-1, fl)) {
	cerr << "ChanSearch: Can't find the " << frame_type 
	     << " frame containing gps " << stop-1 << endl;
	return 2;
    }
    gps_type last_frame = fl.frame_start(stop-1);
    topconf.getChanList(frame_type, last_frame);
    if (stack.back() == topconf) {
	topconf.merge(stack.back());
	stack.pop_back();
    }

    //----------------------------------  Loop until the stack is empty
    while (!stack.empty()) {
	cout << "bot: "   << stack.front().range() 
	     << " prev: " << stack.back().range()
	     << " top: "  << topconf.range() 
	     << endl;

	//------------------------------  Get the mid-point, 
	gps_type mid = (stack.back().stop() + topconf.start()) / 2;
	if (mid >= topconf.start()) {

	    //--------------------------  Add gaps to the end of the stack
	    if (topconf.empty()) {
		stack.back().merge(topconf);
	    }

	    //--------------------------  Write real segments.
	    else {
		if (gzip) topconf.set_compress(chan_config::cls_gzip);
                topconf.write(cacheDir);
	    }

	    //--------------------------  pop off the top entry.
	    topconf = stack.back();
	    stack.pop_back();
	    continue;
	}

	//------------------------------  Test for a match one side or the other
	chan_config test;
	if (debug) test.set_debug(debug);
	test.getChanList(frame_type, mid);

	//------------------------------  Match top?
	if (test == topconf) {
	    topconf.merge(test);
	}

	//------------------------------  No, Match end of stack?
	else if (test == stack.back()) {
	    stack.back().merge(test);
	}

	//------------------------------  No, push it on the stack
	else {
	    stack.push_back(test);
	}
    }
    if (gzip) topconf.set_compress(chan_config::cls_gzip);
    topconf.write(cacheDir);
}
